/* 
 * Copyright (c) 2003 Shaven Puppy Ltd
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are 
 * met:
 * 
 * * Redistributions of source code must retain the above copyright 
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of 'Shaven Puppy' nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */package com.shavenpuppy.jglib.algorithms;

import java.util.*;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Implements the basic heuristic A* route finding algorithm
 */
public class AStar implements PathFinder {

	/** The maximum number of nodes that this AStar can traverse. */
	private final int maxNodes;

	/** The current list of open nodes */
	private final NodeList openList;

	/** The current list of closed nodes */
	private final NodeList closedList;

	/** The map */
	private final Map map;
	
	/** A current list of neighbours. Optimized for 2D grids to begin with but it will adjust itself */
	private ArrayList neighbours = new ArrayList(8);

	/**
	 * A class containing a sorted list of nodes.
	 */
	private static class NodeList {
		final Node[] node;
		int count;
		boolean f_sorted;
		HashMap stateMap = new HashMap();

		NodeList(int size) {
			node = new Node[size];
		}

		// Add a node. Returns false if there's no room.
		boolean addNode(Node newNode) {
			if (count == node.length)
				return false;
			f_sorted = false;
			node[count ++] = newNode;
			
			return true;
		}

		// Remove a node
		void removeNode(State userState) {
			int index = findNode(userState);
			assert index != -1;
			f_sorted = false;
			node[index] = node[--count];
			node[count] = null;
		}

		// Find a node's position by its userstate. Returns -1 if node is not found.
		int findNode(State userState) {
			
			for (int i = 0; i < count; i ++)
				if (node[i].userState.equals(userState))
					return i;
					
			return -1;
		}
		
		// Return a node by its userstate (or null if not present)
		Node getNode(State userState) {
			int index = findNode(userState);
			if (index == -1)
				return null;
			else
				return node[index];
		}

		// Empty the list
		void clear() {
			for (int i = 0; i < count; i ++)
				node[i] = null;
			count = 0;
		}

		// How big is the list?
		int size() {
			return count;
		}
		
		// Is the list empty?
		boolean isEmpty() {
			return count == 0;
		}
		
		// Pop and return the head of the list
		Node popHead() {
			if (count > 0 && !f_sorted) {
				Arrays.sort(node, 0, count);
				f_sorted = true;
			}
			Node ret = node[0];
			node[0] = node[--count];
			node[count] = null;
			return ret;
		}

	}

	/**
	 * Our nodes
	 */
	private static class Node implements Comparable {

		Node parent; // used during the search to record the parent of successor nodes
		Node child; // used after the search for the application to view the search in reverse

		int g; // cost of this node + it's predecessors
		int h; // heuristic estimate of distance to goal
		int f; // sum of cumulative cost of predecessors and self and heuristic

		State userState;

		Node(State userState, Node parent) {
			this.userState = userState;
			this.parent = parent;
		}

		public boolean equals(Object obj) {
			return ((Node) obj).userState.equals(this.userState);
		}

		/**
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public int compareTo(Object o) {
			Node n2 = (Node) o;
			if (this == n2)
				return 0;
			if (n2 == null)
				return 1;
			return f > n2.f ? 1 : f == n2.f ? 0 : -1;
		}
		
		/**
		 * Duplicate the incoming node
		 */
		void dup(Node src) {
			this.g = src.g;
			this.h = src.h;
			this.f = src.f;
		}

	}

	/** The goal node */
	private Node end;
	
	/** The initial node */
	private Node start;

	/** The number of steps */
	private int steps;
	
	/** The current path */
	private ArrayList path;

	/** The current search state */
	private int state = SEARCH_STATE_NOT_INITIALIZED;

	/**
	 * Constructor for AStar.
	 * @param map The map to traverse
	 * @param maxNodes The maximum number of nodes allowed. (For a grid shaped map,
	 * use width * height)
	 */
	public AStar(Map map, int maxNodes) {
		this.map = map;
		this.maxNodes = maxNodes;
		openList = new NodeList(maxNodes);
		closedList = new NodeList(maxNodes);
	}

	/**
	 * @see com.shavenpuppy.jglib.algorithms.PathFinder#findPath(int, int, IntList)
	 */
	public synchronized void findPath(State startState, State endState, ArrayList path) {
		cleanup();
		steps = 0;
		this.path = path;
		state = SEARCH_STATE_SEARCHING;
		start = new Node(startState, null);
		openList.addNode(start);
		start.g = 0;
		start.h = map.getDistance(startState, endState);
		start.f = start.h;
		end = new Node(endState, null);
	}

	/**
	 * Cancel the current search
	 */
	public synchronized void cancel() {
		state = SEARCH_STATE_CANCELLED;
		cleanup();
	}
	
	/**
	 * Clean up when a search is finished
	 */
	private void cleanup() {
		openList.clear();
		closedList.clear();
		neighbours.clear();
	}

	/**
	 * @see com.shavenpuppy.jglib.algorithms.PathFinder#nextStep()
	 */
	public synchronized int nextStep() {

		if (state == SEARCH_STATE_FAILED || state == SEARCH_STATE_SUCCEEDED || state == SEARCH_STATE_CANCELLED)
			return state;

		assert state == SEARCH_STATE_SEARCHING;

		// Failure is defined as emptying the open list as there is nothing left to 
		// search...
		if (openList.isEmpty()) {
			state = SEARCH_STATE_FAILED;
			cleanup();
			return state;
		}

		// Incremement step count
		steps++;

		// Pop the best node (the one with the lowest f) : this is at the head of the list
		Node n = openList.popHead(); // get pointer to the node
		
		// Check for the goal, once we pop that we're done
		if (n.equals(end))
			goalFound(n);
		else
			goalNotFound(n);

		return state;
	}
	
	/**
	 * The goal has been found.
	 * @param n The current node
	 */
	private void goalFound(Node n) {			
		
		//System.out.println("GOAL FOUND in "+steps+" steps!");
		
		end.parent = n.parent;
		
		// A special case is that the goal was passed in as the start state
		// so handle that here
		path.clear();
		if (n != start) {
			// set the child pointers in each node (except Goal which has no child)
			Node nodeChild = n;
			Node nodeParent = n.parent;

			do {
				nodeParent.child = nodeChild;
				nodeChild = nodeParent;
				nodeParent = nodeParent.parent;

			} while (!nodeChild.equals(start)); // Start is always the first node by definition

			// Fill in the path
			n = start;
			do {
				n = n.child;
				if (n != null)
					path.add(n.userState);
			} while (!n.equals(end));
		}

		state = SEARCH_STATE_SUCCEEDED;
		cleanup();
	}
	
	/**
	 * The goal has not been found.
	 * @param n The current node
	 */
	private void goalNotFound(Node n) {

		// We now need to generate the neighbours of this node. We ask the map for the neighbours.
		// By passing in the parent of the current node we help the map to avoid returning a neighbour
		// that simply backtracks.
		
		map.getNeighbours(n.userState, n.parent != null ? n.parent.userState : null, neighbours);

		int numNeighbours = neighbours.size();
		for (int i = 0; i < numNeighbours; i ++) {
			State newState = (State) neighbours.get(i);
			int newg = n.g + map.getCost(n.userState, newState);
			
			// Now find the node on the open or closed lists. If it is on a list
			// already but the node that is already there has a better (lower) g
			// scrore then forget about this neighbour.
			
			// First the open list:
			Node foundOnOpenList = openList.getNode(newState);
			if (foundOnOpenList != null) {
				if (foundOnOpenList.g <= newg) {
					// Already got a node that's cheaper on the open list
					continue;
				}
			}
			
			// Then the closed list:
			Node foundOnClosedList = closedList.getNode(newState);
			if (foundOnClosedList != null) {
				if (foundOnClosedList.g <= newg) {
					// Already got a node that's cheaper on the closed list
					continue;
				}
			}
			
			// This node is the best so far with this particular state, so let's keep it

			// If it was on the closed list, remove it
			if (foundOnClosedList != null)
				closedList.removeNode(foundOnClosedList.userState);
			
			// If it was on the open list, remove it
			if (foundOnOpenList != null)
				openList.removeNode(foundOnOpenList.userState);
				
			// Now add a new node
			Node newNode = new Node(newState, n);
			newNode.g = newg;
			newNode.h = map.getDistance(newState, end.userState);
			newNode.f = newNode.g + newNode.h;
			if (!openList.addNode(newNode)) {
				state = SEARCH_STATE_FAILED;
				cleanup();
				break;
			}
		}

		// push n onto Closed, as we have expanded it now
		if (!closedList.addNode(n)) {
			state = SEARCH_STATE_FAILED;
			cleanup();
		}

	}
}
