package jwars;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.URL;
import java.util.Enumeration;
import javax.imageio.ImageIO;

import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Transparency;
import java.awt.image.BufferedImage;

import jwars.test.JWProperties;
import jwars.model.WarModel;
import jwars.synch.Timer;

/**
 * This class contains miscellaneous static data and methods useful
 * throughout JWars.
 */
public final class Util{

  //Basic program information
  public final static String
      GAME_NAME = "JWars",
      FULL_GAME_NAME = "JWars Warfare Simulation",
      VERSION = "alpha 3 build 2",
      NAME_AND_VERSION = GAME_NAME + " " + VERSION,
      AUTHORS = "Michael Francker Christensen & Ask Hjorth Larsen";

  //private final static String DEFAULT_PROPERTY_FILE = "props.conf";
  public final static String DEFAULT_SERVER_PROPERTY_FILE = "server.conf";
  public final static String DEFAULT_CLIENT_PROPERTY_FILE = "client.conf";

  public final static JWProperties SV_PROPS = JWProperties.loadServerProperties(DEFAULT_SERVER_PROPERTY_FILE);
  public final static JWProperties CL_PROPS = JWProperties.loadClientProperties(DEFAULT_CLIENT_PROPERTY_FILE);

  /**
   * Sets the user name of the local user. This method has any effect only if
   * invoked before networking is initiated; later name changes should be
   * performed using the appropriate networking instructions.
   *
   * @param userName String
   */
  public static void setUserName(String userName){
    CL_PROPS.setProperty(JWProperties.Client.name.name(), userName);
  }

  /**
   * Returns the user name of the local user.
   * @return String
   */
  public static String getUserName(){
    return CL_PROPS.getProperty(JWProperties.Client.name.name());
  }

  //Mimics System.out, though we sadly cannot use final modifier
  public static PrintWriter out = new PrintWriter(System.out);

  //The Timer used for fixing the framerate of JWars
  public final static Timer TIMER = new Timer(WarModel.TIME_PER_FRAME);

  private static boolean drawDebugGrids = false;

  public final static boolean drawDebugGrids(){
    return drawDebugGrids;
  }

  public final static void setDrawDebugGrids(boolean value){
    drawDebugGrids = value;
  }

  /**
   * Redirect the JWars printing facility
   * @param out PrintWriter
   */
  public static void setOut(PrintWriter out){
    Util.out = out;
  }

  //Port used for networking
  public final static int getPort(){
    return Integer.parseInt(CL_PROPS.getProperty(JWProperties.Client.port.name()));
  }

  //Display settings
  public final static GraphicsDevice GRAPH_DEVICE = GraphicsEnvironment.
      getLocalGraphicsEnvironment().getDefaultScreenDevice();

  public final static GraphicsConfiguration GRAPH_CONFIG =
      GRAPH_DEVICE.getDefaultConfiguration();

  static boolean isFullScreen = false;

  /**
   * Sets whether the game should be run in fullscreen. The graphical mode can
   * presently not be changed while the game is running.
   *
   * @param value boolean
   */
  public static void setFullScreen(boolean value){
    isFullScreen = value;
  }

  /**
   * Returns whether the game should be run in fullscreen.
   * @return boolean
   */
  public static boolean isFullScreen(){
    return isFullScreen;
  }

  //private final static int
  //    DEFAULT_WINDOWED_DISPLAY_SIZE_X = 800,
  //    DEFAULT_WINDOWED_DISPLAY_SIZE_Y = 600;

  public final static int DISPLAY_SIZE_X(){
    return isFullScreen() ?
        GRAPH_DEVICE.getDisplayMode().getWidth() :
        Integer.parseInt(CL_PROPS.getProperty(JWProperties.Client.resolution.name()).split("x")[0]);
  }

  public final static int DISPLAY_SIZE_Y(){
    return isFullScreen() ?
        GRAPH_DEVICE.getDisplayMode().getHeight() :
        Integer.parseInt(CL_PROPS.getProperty(JWProperties.Client.resolution.name()).split("x")[1]);
  }

  /**
   * Creates a hopefully managed, opaque image of specified dimensions
   * @param width int
   * @param height int
   * @return BufferedImage
   */
  public static BufferedImage createImage(int width, int height){
    return GRAPH_CONFIG.createCompatibleImage(width, height,
        Transparency.OPAQUE);
  }

  /**
   * Creates a hopefully managed image of specified dimensions and transparency.
   * @param width int
   * @param height int
   * @param transparency int
   * @return BufferedImage
   */
  public static BufferedImage createImage(int width, int height,
      int transparency){
    return GRAPH_CONFIG.createCompatibleImage(width, height, transparency);
  }

  /**
   * Load a bitmasked image from the file denoted by argument String,
   * relative to the top-level JWars directory
   * @param path String
   * @return BufferedImage
   */
  public static BufferedImage readImage(String path){
    return readImage(path, Transparency.BITMASK);
  }

  /**
   * Loads an image with specified transparency from the file denoted by
   * argument String, relative to the top-level JWars directory
   * @param path String
   * @param transparency int
   * @return BufferedImage
   */
  public static BufferedImage readImage(String path, int transparency){
    BufferedImage src = null;
    BufferedImage dst = null;
    String resource = "/graphics/" + path;
    URL url = new Object().getClass().getResource(resource);
    if(url == null){
      System.err.println("<Util> ***WARNING***");
      System.err.println("<Util> Image resource \"" + resource + "\" could not be loaded. You probably forgot to include the graphics directory... AGAIN!");
      System.err.println("<Util> Returning EMPTY IMAGE instead.");
      return createImage(1, 1, Transparency.BITMASK);
    }
    try{
      src = ImageIO.read(url);
    } catch(IOException ex){
      ex.printStackTrace();
    }
    dst = createImage(src.getWidth(), src.getHeight(), transparency);
    Graphics2D g = dst.createGraphics();
    g.drawImage(src, 0, 0, null);
    g.dispose();
    return dst;
  }

  /**
   * Creates a new, hopefully managed BufferedImage identical to the argument
   * (possibly not managed) BufferedImage. Images that have had their data
   * modified directly are generally not managed.
   * @param image Image
   * @return BufferedImage
   */
  public static BufferedImage recreate(BufferedImage image){
    BufferedImage dst = createImage(image.getWidth(null), image.getHeight(null),
        image.getTransparency());
    Graphics2D g2d = dst.createGraphics();
    g2d.drawImage(image, 0, 0, null);
    g2d.dispose();
    return dst;
  }

  /**
   * Returns the angle between minus pi and pi which represents the same
   * direction as the argument.
   *
   * @param angle double
   * @return double
   */
  public static double minusPiToPi(double angle){
    double pi2 = 2 * Math.PI;
    if(angle >= Math.PI || angle <= -Math.PI){
      angle = angle % (pi2);
      if(angle < -Math.PI){
        angle += pi2;
      }
      if(angle > Math.PI){
        angle -= pi2;
      }
    }
    return angle;
  }

  /**
   * Attempts to find the IP address of the local machine.
   *
   * @return the first IP address found, or an empty String if none found.
   */
  public static String getLocalIPAddress(){
    try{
      Enumeration interfaces = NetworkInterface.getNetworkInterfaces();
      while(interfaces.hasMoreElements()){
        NetworkInterface ni = (NetworkInterface)interfaces.nextElement();
        Enumeration addresses = ni.getInetAddresses();
        while(addresses.hasMoreElements()){
          InetAddress ip = (InetAddress)addresses.nextElement();
          //No 127.0.0.1 and no MAC addresses
          if(!ip.isLoopbackAddress() && ip.getHostAddress().indexOf(":") == -1){
            return ip.getHostAddress();
          }
        }
      }
    } catch(SocketException ex){
      ex.printStackTrace();
    }
    return "";
  }

  /**
   * Returns the squared distance between (x1,y1) and (x2,y2).
   *
   * @param x1 int
   * @param y1 int
   * @param x2 int
   * @param y2 int
   * @return the distance
   */
  public static float distSqr(int x1, int y1, int x2, int y2){
    float dx = x2-x1;
    float dy = y2-y1;
    return dx*dx + dy*dy;
  }

  /**
   * Some ATI linux drivers are very bad; this boolean shuts off window
   * decorations which eliminates some bugginess.
   */
  public static boolean useSpecialATIHack = false;

  final static int MEANING_OF_LIFE_THE_UNIVERSE_AND_EVERYTHING = 42;

  private Util(){}

}
