package jwars.data;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Hashtable;
import java.util.InputMismatchException;
import java.util.Scanner;

public class Loader{

  private Hashtable<String, UnitCategory> units =
      new Hashtable<String, UnitCategory>();
  private Hashtable<String, WeaponCategory> weapons =
      new Hashtable<String, WeaponCategory>();
  private Hashtable<String, HullCategory> hullTypes =
      new Hashtable<String, HullCategory>();
  private Hashtable<String, TurretCategory> turretTypes =
      new Hashtable<String, TurretCategory>();

  private Hashtable<String, Reader> readers = new Hashtable<String,Reader>();

  private Hashtable<Character, Double> unitsOfMeasure = new Hashtable<Character,Double>();
  private final static char divisionChar = '/';
  private final static String division = ""+divisionChar;

  private String dataDir = "";
  private String loadCommand = "load";

  private Object lastSuccessfullyLoadedObject = null;

  public Loader(double meter, double second){
    unitsOfMeasure.put('m', meter);
    unitsOfMeasure.put('s', second);
    unitsOfMeasure.put('k', 1000d);
    unitsOfMeasure.put('h', 3600*second);
    initReaders();
  }

  public UnitCategory getUnit(String name){
    UnitCategory cat = units.get(name);
    if(cat == null){
      throw new IllegalArgumentException("No such unit: "+name);
    }
    return cat;
  }

  public WeaponCategory getWeapon(String name){
    WeaponCategory cat = weapons.get(name);
    if(cat == null){
      throw new IllegalArgumentException("No such weapon: "+name);
    }
    return cat;
  }

  public HullCategory getHullType(String name){
    HullCategory cat = hullTypes.get(name);
    if(cat == null){
      throw new IllegalArgumentException("No such hull type: "+name);
    }
    return cat;
  }

  public TurretCategory getTurretType(String name){
    TurretCategory cat = turretTypes.get(name);
    if(cat == null){
      throw new IllegalArgumentException("No such turret: "+name);
    }
    return cat;
  }

  public void load(String path){
    //LineNumberReader lineCounter = new LineNumberReader(new InputStreamReader(
    //  getClass().getResourceAsStream(path)));
    /*try{
      while(true){
        String line = lineCounter.readLine();
        System.out.println(lineCounter.getLineNumber() + " " +
                           line);
      }
    } catch(Exception ex){
      ex.printStackTrace();
    }*/


    //System.exit(0);
    //Scanner scanner = new Scanner(lineCounter);
    Scanner scanner = new Scanner(new BufferedReader(new InputStreamReader(getClass().getResourceAsStream(dataDir+path))));
    try{
      load(scanner);
    } catch(InputMismatchException ex){
      System.err.println("Input mismatch while loading! Last loaded object: "
          +this.lastSuccessfullyLoadedObject);
      System.err.println("File: "+path);
      //System.err.println("Line number: "+lineCounter.getLineNumber());
      System.out.println(""+this);
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public void load(Scanner input) throws InputMismatchException {
    while(input.hasNext()){
      readEntry(input);
    }
  }

  private void readEntry(Scanner input) {
    String nextType = input.next();
    if(loadCommand.equals(nextType)){
      load(input.next());
    } else {
      Reader reader = readers.get(nextType);
      if(reader == null){
        System.err.println("Current contents: " + this);
        throw new IllegalArgumentException("No such category: " + nextType);
      }
      reader.put(input);
    }
  }

  public double readNumber(Scanner input){
    return Double.parseDouble(input.next()) * readUnitsOfMeasurement(input.next());
  }

  private double readUnitsOfMeasurement(String input) {
    double result = 1;
    int i;
    boolean useDivision = false;
    for(i = 0; i < input.length(); i++){
      if(input.charAt(i) == divisionChar){
        if(useDivision){
          throw new IllegalArgumentException("No more than one division sign allowed. Found: "+input);
        }
        useDivision = true;
        continue;
      }
      Double value = unitsOfMeasure.get(input.charAt(i));
      if(value == null){
        throw new IllegalArgumentException("Illegal unit: "+input.charAt(i));
      }
      if(useDivision){
        result /= value;
      } else {
        result *= value;
      }
    }
    return result;
  }

  /**
   * Set the name of the directory, relative to the current working directory,
   * where files are located.
   *
   * @param dir the name of the directory without leading or trailing
   * directory separators
   */
  public void setDataDirectory(String dir){
    this.dataDir = "/"+dir+"/";
  }

  public Hashtable<String, UnitCategory> getUnitTable(){
    return units;
  }

  private void initReaders(){
    new Reader<UnitCategory>(TankCategory.CATEGORY_NAME, units){
      public UnitCategory read(Scanner input){
        return new TankCategory(Loader.this, input);
      }
    };

    new Reader<UnitCategory>(InfantryCategory.CATEGORY_NAME, units){
      public UnitCategory read(Scanner input){
        return new InfantryCategory(Loader.this, input);
      }
    };

    new Reader<WeaponCategory>(WeaponCategory.CATEGORY_NAME, weapons){
      public WeaponCategory read(Scanner input){
        return new WeaponCategory(Loader.this, input);
      }
    };

    new Reader<HullCategory>(HullCategory.CATEGORY_NAME, hullTypes){
      public HullCategory read(Scanner input){
        return new HullCategory(input);
      }
    };

    new Reader<TurretCategory>(TurretCategory.CATEGORY_NAME, turretTypes){
      public TurretCategory read(Scanner input){
        return new TurretCategory(Loader.this,input);
      }
    };

    new Reader<UnitCategory>(FormationCategory.CATEGORY_NAME, units){
      public UnitCategory read(Scanner input){
        return new FormationCategory(Loader.this, input);
      }

    };

    new Reader<UnitCategory>(ArtilleryCategory.CATEGORY_NAME, units){
      public UnitCategory read(Scanner input){
        return new ArtilleryCategory(Loader.this, input);
      }
    };
    new Reader<UnitCategory>(AssaultGunCategory.CATEGORY_NAME, units){
      public UnitCategory read(Scanner input){
        return new AssaultGunCategory(Loader.this, input);
      }
    };
  }

  private abstract class Reader<T> {

    public final Hashtable<String, T> table;

    public Reader(String name, Hashtable<String, T> table){
      this.table = table;
      readers.put(name, this);
    }

    public abstract T read(Scanner input);

    public void put(Scanner input){
      String name = input.next();
      T object = read(input);
      table.put(name, object);
      Loader.this.lastSuccessfullyLoadedObject = object;
    }

    public String toString(){
      return table.toString();
    }

  }

  public static void main(String[] args) throws IOException {
    Scanner s = new Scanner(System.in);
    Loader l = new Loader(1.0, 1.0);
    l.load(s);
  }

  public String toString(){
    StringBuffer s = new StringBuffer();
    s.append("---Loader contents---");
    s.append("\nUnits:\n");
    s.append(units.toString());
    s.append("\nWeapons:\n");
    s.append(weapons.toString());
    s.append("\nHull types:\n");
    s.append(hullTypes.toString());
    s.append("\nTurret types:\n");
    s.append(turretTypes.toString());
    s.append("\n");
    return s.toString();
  }

}
