package jwars.model;

import jwars.data.FormationCategory;
import jwars.model.ai.impl.FormationAI;

/**
 * A Formation represents a Unit which is strictly an organization into
 * sub units. Every non-empty Formation a leader, which is selected among its
 * sub units. The location of this Formation is defined to be equal to the
 * location of its leader.
 */
public class Formation
    extends Unit{

  //Used to make sure that units are added without overlap
  //private final static float unitSeparationFactor = 1.4f;

  //A formation will be considered centered on its leader.
  private Unit leader;

  //private double theta;

  public Formation(FormationCategory cat){
    this(cat.name, cat.radius);
  }

  public Formation(String name, int radius){
    super(name, radius);
    this.setAI(new FormationAI(this));
  }

  /**
   * Adds all sub Units of this Unit to the world.
   *
   * @param world World
   * @param x int
   * @param y int
   */
  protected void addChildrenToWorld(){
    for(int i = 0; i < subUnits.size(); i++){
      //x += unitAdditionSeparationFactor * u.radius;
      Unit u = subUnits.get(i);
      if(u.alive){
        int x = (int)DefaultFormationPositions.getX(i, theta, u.radius);
        int y = (int)DefaultFormationPositions.getY(i, theta, u.radius);
        u.location.set(location.x + x,location.y + y);
        u.setAngle(this.getAngle());
        //u.addToWorld(location.x + x, location.y + y, true);
      }
    }
    super.addChildrenToWorld();
  }

  public void setAngle(double theta){
    super.setAngle(theta);
  }

  /*public void update(){
    super.update();
    ((FormationAI)controller).setLocation(location.x, location.y);
  }*/

  /**
   * This method is called when no sub units are left.
   */
  public void reportAllUnitsDestroyed(){
  }

  /**
   * Adds argument Unit to this Formation. If this Formation does not yet
   * have any sub Units, that Unit is now the leader of this Formation.
   * @param unit Unit
   */
  public void addUnit(Unit unit){
    super.addUnit(unit);
    if(this.leader == null){
      setLeader(unit);
    }
  }

  public void setLeader(Unit unit){
    if(unit.superUnit != this){
      throw new IllegalArgumentException(
          "Leader must be a sub unit of this Formation!");
    }
    this.leader = unit;
    this.location = leader.location;
  }

  /**
   * Will check whether the removed Unit is this Formation's leader,
   * and if so attempt to find a new leader among onboard elements of this
   * Formation. If no onboard units exist, this Formation is removed from the
   * world (i.e. considered offboard) and a leader will be selected among the
   * offboard elements of this Formation.
   *
   * If this fails, the Formation
   * has no remaining elements and the reportAllUnitsDestroyed method will
   * be invoked.
   *
   * If this method is overridden, it should be invoked by the overriding
   * method.
   *
   * @param u Unit which has been removed
   */
  protected void reportSubUnitDestroyed(Unit u){
    super.reportSubUnitDestroyed(u);
    if(!isAlive()){
      return;
    }
    if(u == this.leader){
      Unit newLeader = null;
      newLeader = findNewLeader(true); //Find onboard leader if possible
      if(newLeader == null){
        if(this.onBoard)
          this.removeFromWorld(); //No onboard units left
        newLeader = findNewLeader(false); //Find an offboard one, then
      }
      if(newLeader == null){ //No offboard units left either
        reportAllUnitsDestroyed();
      } else{
        setLeader(newLeader);
      }
    }
  }

  /**
   * Finds a new leader for this Unit as described in the reportSubUnitRemoved
   * documentation.
   *
   * @param mustBeOnboard boolean whether the desired leader must be onboard
   * @return Unit the new leader, or null if no leader is found
   */
  private Unit findNewLeader(boolean mustBeOnboard){
    Unit newLeader = null, leaderCandidate = null;
    for(int i = 0; (newLeader == null) && (i < subUnits.size()); i++){
      leaderCandidate = subUnits.get(i);
      if(leaderCandidate.alive && (mustBeOnboard || leaderCandidate.onBoard)){
        newLeader = leaderCandidate;
      }
    }
    return newLeader;
  }

}
