package jwars.model;

import java.awt.Graphics2D;

import static jwars.Util.minusPiToPi;
import jwars.data.MoveableCategory;
import jwars.gui.units.Sprite;
import jwars.model.ai.impl.MoveableAI;
import jwars.model.weapons.WeaponModel;
import jwars.model.vision.Observer;
import jwars.model.vision.ObserverModel;
import jwars.model.collision.*;
import jwars.model.weapons.ArmourModel;
import jwars.model.weapons.Damageable;

/**
 * Moveables represent physical entities which can move around, such as
 * vehicles or infantry squads. Units
 */
public class Moveable extends Unit implements Collider, Observer, Damageable {

  public final static double dt = WarModel.TIME_PER_FRAME/1000d;
  public final static double second = WarModel.second;
  public final static double meter = WarModel.meter;

  public final static double kilometer = 1000*meter;
  public final static double minute = 60*second;
  public final static double hour = 60*minute;

  private final static Sprite EMPTY_SPRITE = new EmptySprite();

  private final CollisionProperties collisionProps;
  private final ObserverModel obsModel;
  //private final ArmourModel armour;

  public final MoveableCategory CAT;

  //The orientation of this Unit. Measured in radians.
  //protected double theta;
  //Current and maximum speed of this Unit. This will be replaced by a model
  //supporting different terrain types later.
  protected double speed;
  //protected double turnSpeed = 2*Math.PI;
  private Sprite sprite;
  private MoveableAI controller; //The AI of this Moveable with proper type
  //We should provide a proper identifier at some time.

  protected final WeaponModel weaponry;// = new WeaponModel();

  /**
   * Construct a unit
   * @param radius radius of this Unit, measured in units, for collision detection
   * @param sprite Sprite responsible for rendering this unit
   * @param x x-value of this Unit's location
   * @param y y-value of this Units location
   * @param theta angle of this Unit, measured in radians, with respect to the
   * x-axis
   * @param maxSpeed Maximum speed this Unit is capable of moving.
   */
  public Moveable(MoveableCategory cat,/*, String name, int radius, double maxSpeed,*/ boolean massive){
    super(cat.name, cat.radius);
    this.CAT = cat;
    //this.armour = CAT.armour.create();
    controller = new MoveableAI(this);
    super.setAI(controller);
    setSprite(EMPTY_SPRITE);
    weaponry = new WeaponModel(controller, cat.weaponry);
    //this.maxSpeed = maxSpeed;
    this.sprite = sprite;
    this.obsModel = new ObserverModel(WarModel.WORLD.obsEnvironment, this);
    this.collisionProps = new CollisionProperties(this, massive, true);
  }

  public void setLocation(int x, int y){
    location.set(x,y);
  }

  public ArmourModel getArmourModel(){
    return ArmourModel.NULL_ARMOUR_MODEL;
  }

  /**
   * Update the state of this Unit
   */
  public void update(){
    super.update();
    if(controller.isFiring()){
      updateFiring();
    }
    if(controller.isMoving()){
      move();
    }
  }

  public void updateFiring(){
    double targetAngle = controller.getTargetAngle();
    boolean canFire = aimTowards(targetAngle) && weaponry.canFire();
    if(canFire){
      weaponry.fire(getX(), getY(), controller.getTargetX(), controller.getTargetY()); //Make this method more sensible
    }
  }

  /**
   * Move this Unit according to destination and speed. This method should be
   * invoked only from the Unit move queue.
   */
  public void move(){
    double newTheta = minusPiToPi(controller.getMoveAngle());
    double angleDifference = minusPiToPi(newTheta - this.theta);
    double maxRotation = CAT.turnRate * dt;
    if(Math.abs(angleDifference) > maxRotation){
      this.theta += maxRotation * Math.signum(angleDifference);
      this.dirtify();
      return;
    } else {
      this.theta = newTheta;
    }

    speed = CAT.speed;
    int dx = (int)(speed*Math.cos(theta)*dt);
    int dy = (int)(speed*Math.sin(theta)*dt);
    int newX = location.x + dx;
    int newY = location.y + dy;

    boolean hasMoved = world.DETECTOR.move(this,newX,newY);
    if(hasMoved){
      this.theta = theta; //Only actually turn if the move was successful
      controller.closeEnough((int)(speed * dt));
    }
  }

  /**
   * Add this Moveable to the world and register it in the collision detector.
   * @param world World
   * @param x int
   * @param y int
   */
  public void addToWorld(int x, int y, boolean addChildren){
    super.addToWorld(x, y, addChildren);
    world.DETECTOR.register(this);
    weaponry.setFireAtWill(true);
    world.obsEnvironment.register(obsModel);
    dirtify();
  }

  public void removeFromWorld(){
    world.DETECTOR.remove(this);
    world.obsEnvironment.remove(obsModel);
    controller.setDestination(0,0);
    controller.orderHoldFire();
    weaponry.setFireAtWill(false);
    super.removeFromWorld();
  }

  /**
   * This method will throw an UnsupportedOperationException. It is considered
   * illegal to add sub Units to Moveables, being the lowest level of
   * organization.
   *
   * @param u
   */
  public void addUnit(Unit u){
    throw new UnsupportedOperationException("Moveables cannot have sub units");
  }

  public boolean aimTowards(double theta){
    return true;
  }

  /**
   * Retrieve the angle of this Moveable in the main coordinate system.
   * @return double the angle
   */
  public double getAngle(){
    return theta;
  }

  /**
   * Alert collision detector and sprite rendering mechanisms of this Moveable
   * having changed.
   */
  public void dirtify(){
    if(!isAlive()){
      return;
    }
    sprite.setDirty(true);
    world.DETECTOR.dirtify(this);
  }

  /**
   * Retrieve the Sprite associated with this Moveable.
   * @return Sprite
   */
  public Sprite getSprite(){
    return sprite;
  }

  /**
   * Set the Sprite of this Moveable
   * @param sprite the new Sprite
   */
  public void setSprite(Sprite sprite){
    this.sprite = sprite;
  }

  /**
   * Fills the argument int[] of length at least 4 with the
   * bounding box data of this Moveable.
   *
   * @param storage int[]
   * @return int[]
   */
  public int[] getBounds(int[] storage){
    storage[0] = location.x-radius;
    storage[1] = location.y-radius;
    storage[2] = location.x+radius;
    storage[3] = location.y+radius;
    return storage;
  }

  public void destroyLater(){
    world.destroyLater(this);
  }

  public void collisionDetected(Collider other){
    controller.walkAround(other);
  }

  public CollisionProperties getCollisionProps(){
    return this.collisionProps;
  }

  public ObserverModel getObserverModel(){
    return this.obsModel;
  }

  public WeaponModel getWeaponModel(){
    return this.weaponry;
  }

  public boolean sharesVision(Observer other){
    Moveable m = (Moveable)other; //Ugly cast
    return !getTeam().enemyOf(m.getTeam());
  }

  public boolean visionCheck(Observer other, int dist, int timeIncrement){
    //Will only be invoked if within max visibility range of other, so
    //just return true to achieve circular sight field
    boolean hasBeenSpotted = true;
    if(hasBeenSpotted){ //Possibly set new target
      Moveable target = (Moveable)getAI().getTarget();
      if(target == null){
        controller.orderAttack((Moveable)other);
      } else { //check whether target is closer than other
        float oldDistSqr = jwars.Util.distSqr(location.x, location.y,
            target.getX(),
            target.getY());
        float newDistSqr = (float)dist * dist;
        if(newDistSqr < oldDistSqr){
          controller.orderAttack((Moveable)other);
        }
      }
    }
    return hasBeenSpotted;
  }

  /**
   * Returns whether this unit has been spotted.
   *
   * @param other Observer
   * @return boolean
   */
  public boolean hasBeenSpotted(Observer other){
    return obsModel.hasBeenSpotted();
  }

  /*public void changeCollisionProperties(boolean massive, boolean mobile){
    collisionProps = new CollisionProperties(massive, mobile);
  }*/

  /**
   * This is the default Sprite implementation which is invisible. Units that
   * are visible should provide their own Sprite implementations.
   */
  private static class EmptySprite extends Sprite{

    private EmptySprite(){
      super(0);
    }

    public void draw(Graphics2D g2d, int x, int y){
    }

  }

}
