package jwars.model;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import jwars.event.dynamic.Schedulator;
import jwars.model.collision.CollisionDetector;
import jwars.model.collision.CollisionTile;
import jwars.model.map.CoordinateSystem;
import jwars.model.map.StructureFactory;
import jwars.model.map.TerrainTile;
import jwars.model.map.TileMap;
import jwars.model.vision.ObservationEnvironment;

/**
 * This class contains the state of the world, i.e. collision- and terrain
 * tilemaps, and stores all units in the game sorted by formations.
 */
public class World{

  public final CoordinateSystem MAIN_COORDS;
  public final TileMap<CollisionTile> COLLISION_MAP;
  public final TileMap<TerrainTile> TERRAIN_MAP;

  public final CollisionDetector DETECTOR;

  private final Unit EVERYONE = Unit.createContainerUnit(this);
  private final List<WorldListener> listeners = new ArrayList<WorldListener>();

  public final double meter;
  public final double second;

  public final ObservationEnvironment obsEnvironment;

  //private final EventQueue overwatchQueue = new EventQueue(10);
  private DynamicSprite dynSprite = null;
  private Schedulator schedule = new Schedulator();

  /**
   * This is the centralized random number generator of this world.
   * This generator should be used ONLY to generate data that is generated
   * on ALL clients, otherwise game will go out of synch.
   *
   * In other words, every client should make exactly the same number of calls
   * at exactly the same frame counts to this generator.
   *
   * Use another generator to randomize graphical effects or similar.
   */
  public final Random rand = new Random(0);

  //Units designated for removal at the end of this update.
  //Will be purged once per frame.
  private final List<Unit> toBeRemoved = new ArrayList<Unit>();

  public World(CoordinateSystem mainCoords,
               TileMap<CollisionTile> collisionMap,
               TileMap<TerrainTile> terrainMap,
               double meter,
               double second){
    jwars.Util.TIMER.addTimerListener(schedule);
    this.meter = meter;
    this.second = second;
    this.MAIN_COORDS = mainCoords;
    this.COLLISION_MAP = collisionMap;
    this.TERRAIN_MAP = terrainMap;
    this.DETECTOR = new CollisionDetector(this);
    this.obsEnvironment = new ObservationEnvironment(jwars.Util.TIMER, MAIN_COORDS, 50, 250, metric(500));
    EVERYONE.addToWorld(0, 0, false);
    int width = (int)(MAIN_COORDS.WIDTH/WarModel.meter);
    int height = (int)(MAIN_COORDS.HEIGHT/WarModel.meter);
    System.out.println("World initialization completed. World size: "+width+"m times "+height+"m.");
  }

  //Units will be enumerated according to these tree
  private static int bitsPerNestingLevel = 5;
  private static int maxUnits = (1<<bitsPerNestingLevel) - 1;
  private static int treeDepth = 6;
  //30 of the 32 bits of an int are used to enumerate units


  /**
   * Fetch the Unit which has the argument ID
   * @param id ID of Unit to be fetched
   * @return Unit Unit with argument ID
   */
  public Unit getUnit(int id){
    Unit u = EVERYONE;
    for(int i = treeDepth; i >= 0; --i){
      u = u.getUnit(((id >>> (i * bitsPerNestingLevel)) & maxUnits) - 1);
    }
    return u;
  }

  /**
   * Used to set the initial conditions not provided by constructor.
   * For example, add units, set their objectives, and any other
   * configurations.
   */
  public void init(){
    StructureFactory.buildDefaultStructures(this);
    EVERYONE.checkAllIds(this); //debug
    System.out.println("Default world setup loaded");
  }

  void unitRegistered(Unit unit){
    for(WorldListener listener : listeners){
      listener.unitNowOnBoard(unit);
    }
  }

  void unitRemoved(Unit unit){
    for(WorldListener listener : listeners){
      listener.unitNowOffBoard(unit);
    }
  }

  public void destroyLater(Unit m){
    toBeRemoved.add(m);
  }

  public void addDynamicSprite(DynamicSprite sprite){
    sprite.next = dynSprite;
    if(dynSprite != null)
      dynSprite.previous = sprite;
    dynSprite = sprite;
  }

  public void schedule(jwars.event.dynamic.DEvent e, int time){
    schedule.scheduleFor(e, time);
  }

  public void waitFor(jwars.event.dynamic.DEvent e, int delay){
    schedule.schedule(e, delay);
  }

  public DynamicSprite getFirstDynamicSprite(){
    return dynSprite;
  }

  void removeDynamicSprite(DynamicSprite sprite){
    if(sprite.previous == null){
      dynSprite = sprite.next;
    } else {
      sprite.previous.next = sprite.next;
      sprite.next.previous = sprite.previous;
      sprite.previous = null;
      //Don't set sprite.next to null, since that will prevent it from
      //returning the correct object throught a next() call
    }
  }

  public void addTopLevelUnit(Unit unit){
    EVERYONE.addUnit(unit);
    //if(unit instanceof Moveable){
    //  Moveable m = (Moveable)unit;
    //  m.BRAIN.stop();
    //  DETECTOR.register(m);
    //}
  }

  public Unit getEveryOne(){
    return EVERYONE;
  }

  public void update(){
    EVERYONE.update();
    destroyPending();
  }

  private void destroyPending(){
    for(int i = toBeRemoved.size()-1; i >= 0; i--){
      Unit u = toBeRemoved.remove(i);
      if(u.world == this){
        //There may be duplicate entries, so avoid calling removal methods more
        //than once by ignoring those that are not
        //registered with this World
        u.destroy();
      }
    }
    assert(toBeRemoved.isEmpty());
  }

  public void addUnitListener(WorldListener listener){
    listeners.add(listener);
  }

  public void removeUnitListener(WorldListener listener){
    listeners.remove(listener);
  }

  public int relativeToAbsoluteX(double x){
    return (int)(x*MAIN_COORDS.WIDTH);
  }

  public int relativeToAbsoluteY(double y){
    return (int)(y*MAIN_COORDS.HEIGHT);
  }

  public int countTotalUnits(){
    return countSubUnits(EVERYONE);
  }

  public int countSubUnits(Unit u){
    if(!u.isAlive()){
      return 0;
    }
    if(u.getSubUnitCount() == 0){
      return 1;
    } else {
      int currentTotal = 0;
      for(int i = 0; i < u.getSubUnitCount(); i++){
        currentTotal += countSubUnits(u.getUnit(i));
      }
      return currentTotal;
    }
  }

  public int secondsToFrames(double seconds){
    return (int)(1000 * seconds / WarModel.TIME_PER_FRAME);
  }

  public double framesToSeconds(int frameCount){
    return (double)frameCount * WarModel.TIME_PER_FRAME * 1000;
  }

  public double toMetres(int dist){
    return dist / WarModel.meter;
  }

  public int metric(double meters){
    return (int)(meters * WarModel.meter);
  }

}
