package jwars.model.ai.impl;

import jwars.model.Coord;
import jwars.model.Unit;
import jwars.model.ai.ListeningAI;
import jwars.model.ai.MobileAI;
import java.util.Iterator;
import jwars.model.ai.Position;

public abstract class MovementQueueAI
    implements MobileAI{

  private MoveOrder nextOrder; //only used for queueing
  private MoveOrder lastOrder;

  //Immediate destination
  private Coord destination = new Coord();
  //Current location of the entity represented by this AI
  //private Coord location;
  //The AI which has given the current order
  private ListeningAI currentSource;

  private final Unit owner;

  private boolean isMoving = false;

  public MovementQueueAI(Unit u){
    //location = new Coord(x,y);
    this.owner = u;
    //this.location.set(x,y);
  }

  public void orderMove(ListeningAI source, int x, int y){
    //Flush the queue
    nextOrder = null;
    lastOrder = null;
    //Order the move
    moveTowards(x, y);
    currentSource = source;
  }

  public void changeNextWayPoint(ListeningAI source, int x, int y){
    currentSource = source;
    moveTowards(x,y);
  }

  private void moveTowards(int x, int y){
    destination.set(x, y);
    isMoving = true;
    setDestination(x,y);
  }

  public void appendMove(ListeningAI source, int x, int y){
    if(nextOrder == null){ //At most one order being executed already
      if(isMoving){ //One order being executed already
        nextOrder = new MoveOrder(source, x, y);
        lastOrder = nextOrder;
      } else{ //No orders at all being executed right now
        moveTowards(x, y);
        currentSource = source;
      }
    } else{ //Orders are already in the queue
      assert(lastOrder != null);
      lastOrder.next = new MoveOrder(source, x, y);
      lastOrder = lastOrder.next;
    }
  }

  /**
   *
   * @param source ListeningAI
   * @param x int
   * @param y int
   */
  public void prependMove(ListeningAI source, int x, int y){
    if(isMoving){
      //Backup what is presently going on
      MoveOrder next = new MoveOrder(currentSource, destination.x,
          destination.y);
      //Store this new order before the rest
      next.next = nextOrder;
      nextOrder = next;
      //Finally, execute the given order
      moveTowards(x, y);
      currentSource = source;
    }

  }

  /**
   * Sets the destination equal to the location of the entity represented by
   * this AI.
   * @param source ListeningAI
   */
  public void orderHalt(ListeningAI source){
    destination.set(owner.getX(), owner.getY());
    currentSource = null;
    nextOrder = lastOrder = null;
    isMoving = false;
  }

  /**
   * Tests whether the location lies within the provided tolerance of the
   * destination and returns the result. If more move orders are enqueued, they
   * will be executed shortly.
   *
   * @param tolerance the tolerance
   * @return boolean whether location and destination are close enough
   */
  public boolean closeEnough(int tolerance){
    boolean nearEnough = destination.dist(owner.getX(), owner.getY()) <= tolerance;
    if(nearEnough){
      destination.x = owner.getX();
      destination.y = owner.getY();
      pumpOrders();
    }
    return nearEnough;
  }

  public ListeningAI currentAutority(){
    return currentSource;
  }

  public ListeningAI nextAuthority(){
    if(nextOrder != null){
      return nextOrder.source;
    } else {
      return null;
    }
  }

  /**
   * Execute the next MoveOrder in the order queue if non-empty. Otherwise
   * report successful completion of orders.
   */
  public void pumpOrders(){
    if(nextOrder != null){ //There is something still in the queue
      currentSource = nextOrder.source;
      //destination.set(nextOrder.x, nextOrder.y);
      moveTowards(nextOrder.x, nextOrder.y);
      nextOrder = nextOrder.next;
    } else{ //All orders completed
      isMoving = false;
      if(currentSource != null)
        currentSource.reportFinished(this);
    }
  }

  /**
   * Get the angle with respect to the x-axis in which the owner of this AI
   * must move in order to reach the target.
   * @return the angle
   */
  public double getMoveAngle(){
    return Math.atan2(destination.y - owner.getY(), destination.x - owner.getX());
  }

  public boolean isMoving(){
    return isMoving;
  }

  //Hack to make the iterator work properly.
  //Implement something nicer if we have time
  private final MoveOrder currentDestination = new MoveOrder(null, 0, 0){

    public MoveOrder next(){
      return nextOrder;
    }

    public int getX(){
      return destination.x;
    }

    public int getY(){
      return destination.y;
    }

  };

  public Iterator<Position> iterator(){
    return new Iterator<Position>(){

      private MoveOrder next = currentDestination;

      public boolean hasNext(){
        return next != null;
      }

      public Position next(){
        if(next == null){
          throw new java.util.NoSuchElementException();
        }
        MoveOrder returnValue = next;
        next = next.next();
        return returnValue;
      }

      public void remove(){
        throw new UnsupportedOperationException();
      }

    };
  }

  /*public void setLocation(int x, int y){
    location.set(x,y);
  }*/

  /*private void abort(AI source){
    if(source != currentSource && currentSource != null){
      currentSource.reportFailed(this);
    }
    currentSource = source;
     }*/

}
