package jwars.model.map;



public class CoordinateSystem{

  public final int WIDTH;
  public final int HEIGHT;
  final int scale;
  private final CoordinateSystem superSystem;

  /**
   * Creates a CoordinateSystem as a subsystem of argument CoordinateSystem.
   * @param mother the mother of this CoordinateSystem
   * @param scale size of units in this CoordinateSystem measured in
   * the mother CoordinateSystem's units
   */
  public CoordinateSystem(CoordinateSystem superSystem, int scale){
    while(superSystem.superSystem != null){
      scale *= superSystem.scale;
      superSystem = superSystem.superSystem;
    }
    this.scale = scale;
    this.superSystem = superSystem;

    //Presently we don't accept nonzero remainders!
    //if(superSystem.WIDTH % scale != 0)
    //  throw new IllegalArgumentException("Nonzero remainder from division by scale");
    this.WIDTH = superSystem.WIDTH / scale;
    this.HEIGHT = superSystem.HEIGHT / scale;
  }

  /**
   * Creates a CoordinateSystem of specified dimensions
   * @param width number of units in horizontal direction
   * @param height number of units in vertical direction
   */
  public CoordinateSystem(int width, int height){
    this.scale = 1;
    this.superSystem = null;
    this.WIDTH = width;
    this.HEIGHT = height;
  }

  public boolean isLegalXCoordinate(int x){
    return x >= 0 && x < WIDTH;
  }

  public boolean isLegalYCoordinate(int y){
    return y >= 0 && y < HEIGHT;
  }

  /**
   * Creates and returns the largets CoordinateSystem with the specified
   * supersystem and dimensions smaller than maxWidth and maxHeight,
   * respectively.
   * @param superSystem the supersystem of this CoordinateSystem
   * @param maxWidth the maximally allowed width of this CoordinateSystem
   * @param maxHeight the maximally allowed height of this CoordinateSystem
   * @return the created CoordinateSystem
   */
  public static CoordinateSystem createLargestCompatibleSystem(CoordinateSystem
      superSystem, int maxWidth, int maxHeight){
    //Presently only supports square Coordinate Systems
    //int scale = Math.max(superSystem.WIDTH/maxWidth,
    //                     superSystem.HEIGHT/maxHeight);
    int xError = superSystem.WIDTH % maxWidth;
    //maxWidth -= xError;
    while(xError != 0){
      maxWidth--;
      xError = superSystem.WIDTH % maxWidth;
    }
    int yError = superSystem.HEIGHT % maxHeight;
    while(yError != 0){
      maxHeight--;
      yError = superSystem.HEIGHT % maxHeight;
    }

    int scale = Math.max(superSystem.WIDTH / maxWidth,
                         superSystem.HEIGHT / maxHeight);

    CoordinateSystem system = new CoordinateSystem(superSystem, scale);
    return system;
  }

  /**
   * Transforms integer to target coordinate system
   * @param target Coordinate system to transform to
   * @param x value to transform
   * @return value of transformed coordinate
   */
  public int transformToCoords(CoordinateSystem target, int x){
    return x * this.scale / target.scale;
  }

  /**
   * Transforms all elements of argument integer array to argument coordinate
   * system and returns the array.
   * @param target Coordinate system to transform to
   * @param values values to be transformed
   * @return array of transformed values
   */
  public int[] transformToCoords(CoordinateSystem target, int[] values){
    return transformToCoords(target, values, 0, values.length);
  }

  /**
   * Transforms the specified elements of argument integer array to argument
   * coordinate system and returns the array.
   * @param target Coordinate system to transform to
   * @param values container of values to be transformed
   * @param offset index of first element to be transformed
   * @param len number of elements to be transformed
   * @return array containing transformed values
   */
  public int[] transformToCoords(CoordinateSystem target, int[] values, int offset, int len){
    for(int i=offset; i < offset + len; i++){
      values[i] = values[i] * this.scale / target.scale;
    }
    return values;
  }

}
