package jwars.net;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * Class used for sending and receiving instructions or objects according to
 * a user specified protocol. Extending IOHandler<T> signifies that the
 * extending class is capable of encoding and sending an instruction,
 * optionally in the form of an object of type T.
 *
 * Using this class to send and receive data is less bandwidth-intensive than
 * serialization or externalization using the Serializable or Externalizable
 * interfaces of the java.io package. Overhead for sending one instruction of
 * arbitrary type is one byte.
 *
 * IOHandlers must be added to a Protocol in order to function.
 */
public abstract class IOHandler<T>{

  //
  final static int unAssigned = Byte.MIN_VALUE;
  //Each IOHandler registered within a Protocol has its own unique number
  byte uniqueId = unAssigned;
  //This IOHandler will write to the DataOutput specified by its Protocol object
  DataOutputStream output;

  /**
   * Called by a Client upon reception of the ID code
   * associated with this IOHandler. Implement this method to specify the
   * action appropriate on reception.
   *
   * @param input DataInput
   * @param timeStamp int
   * @throws IOException
   */
  protected abstract void read(DataInput input, int timeStamp) throws IOException;

  /**
   * Called by a Server upon reception of the ID code
   * associated with this IOHandler. This method exists because servers
   * will typically handle input differently compared to clients. Specifically,
   * a server will frequently want to echo the data to different clients, but
   * it is up to the implementation to decide whether any modifications should
   * be done to the sent data.
   *
   * @param input DataInputStream
   * @throws IOException
   */
  protected abstract void echo(int clientIdentifier, DataInput input, DataOutput output) throws IOException;

  /**
   * Write the data associated with the argument object of type T to argument
   * DataOutputStream. This method should behave similarly to writeExternal() in
   * interface Externalizable of the java API.
   *
   * The write method should not be invoked directly;
   * use the send method instead.
   *
   * @param t T
   * @param output DataOutputStream
   * @throws IOException
   */
  protected abstract void write(T t, DataOutput output) throws IOException;

  /**
   * Send arguement object of type T using this IOHandler. This method
   * invokes the write method using an encapsulated OutputStream.
   * @param t object to be sent
   * @throws IOException
   */
  public final void send(T t) throws IOException{
    output.writeByte(uniqueId);
    write(t, output);
    output.flush();
  }

  /**
   * Returns a String representation of this IOHandler.
   * @return String
   */
  public String toString(){
    return super.toString() + " ID = "+uniqueId;
  }

}
