package jwars.net;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.List;
import java.util.Vector;

import jwars.synch.Timer;

public class Server{

  //Used to time stamp instructions
  private final Timer timer;
  private final ListeningThread listeningThread = new ListeningThread();
  private final ServerSocket servSock;
  private final List<ClientThread> clients = new Vector<ClientThread>();
  private final Protocol protocol;
  //This stream sends instructions to all clients
  protected final DataOutputStream splitOutput = new DataOutputStream(new
      SplitStream());
  final ServerEventHandler eventHandler;
  private int latency;
  private int nextClientNumber; //Used to enumerate clients
  private int lastReturnedTimeStamp;

  public Server(Protocol protocol, Timer timer, ServerEventHandler eventHandler,
                int port, int latency) throws IOException{
    this.protocol = protocol;
    protocol.server = this;
    this.timer = timer;
    this.latency = latency;
    this.eventHandler = eventHandler;
    this.servSock = new ServerSocket(port);
    listeningThread.start();
  }

  /**
   * Invoked when an exception is thrown as a result of client I/O.
   * This will result in the client being kicked and output being displayed.
   *
   * @param c ClientThread
   * @param ex Exception
   */
  private void reportClientException(ClientThread c, Exception ex){
    System.out.println("Server: Exception in Client Thread " + c+" [id="+c.clientNumber+"] ");
    synchronized(this){
      removeClient(c);
    }
    System.out.println("Offending client has been kicked!");
    jwars.Util.out.println("Client [" + c + "] has left the game!");
  }

  /**
   * Closes the server socket which accepts further connections.
   */
  public void stopListening(){
    listeningThread.isListening = false;
    try{
      servSock.close();
    } catch(IOException ex){
      ex.printStackTrace();
    }
  }

  /**
   * Removes the argument client and notifies the event handler that this has
   * happened.
   *
   * @param c ClientThread
   */
  private synchronized void removeClient(ClientThread c){
    try{
      //Client could be removed already if an exception was thrown while
      //attempting to dislodge it 'nicely'.
      clients.remove(c);
      processInstruction(protocol.clientLeft, c.clientNumber, c.input, splitOutput);
      eventHandler.clientRemoved(c.clientNumber, c.input, c.output, splitOutput);
      splitOutput.flush();
    } catch(IOException ex){
      ex.printStackTrace();
    }
  }

  /**
   * Registers the argument client and notifies event handlers. This will
   * result in the handShake method of the handler being invoked, which may
   * decide to not accept the particular client.
   *
   * @param client ClientThread
   * @throws IOException
   */
  private synchronized void addClient(ClientThread client) throws IOException{
    boolean isAccepted = false;
    //First, add and handshake
    try{
      isAccepted = eventHandler.handShake(client.clientNumber, client.input, client.output);
    } catch(IOException ex){
      System.out.println("Exception thrown during handshake");
    } finally {
      if(!isAccepted){
        System.out.println("Client disconnected");
        client.isListening = false;
        client.input.close();
        client.output.close();
        return;
      }
    }
    clients.add(client);
    System.out.println("Server: New client [id="+client.clientNumber+"] connected");
    //Send the instruction to others that a new client has joined
    processInstruction(protocol.clientJoined, client.clientNumber, client.input, client.output);
    //splitOutput.writeInt(client.clientNumber);
    eventHandler.clientJoined(client.clientNumber, splitOutput);
    splitOutput.flush();
    //Finally activate the client's own thread taking care of further communication
    client.start();
  }

  /**
   * Sets the latency of this Server.
   * @param latency int
   */
  public void setLatency(int latency){
    this.latency = latency;
  }

  /**
   * Returns the latency of this Server.
   * @return int
   */
  public int getLatency(){
    return latency;
  }

  /**
   * Returns the time at which an instruction processed currently would be
   * expected to be executed on clients.
   *
   * @return int
   */
  public int getTimeStamp(){
    return lastReturnedTimeStamp = Math.max(timer.getTime() + latency,
        lastReturnedTimeStamp);
  }

  /**
   * Sends a keep-alive to all clients. This should be done once in a while,
   * possibly around two times per interval of the duration returned by
   * getLatency().
   */
  public synchronized void sendKeepAliveInstruction(){
    try{
      IOHandler instr = protocol.getKeepAliveInstruction();
      splitOutput.writeByte(instr.uniqueId);
      //Perhaps call the keepalive echo
      splitOutput.writeInt(getTimeStamp());
      splitOutput.flush();
    } catch(NullPointerException ex){
      System.err.println("KeepAliveInstruction should not read data from Input"
                         + "Stream in echo method!");
      ex.printStackTrace();
    } catch(IOException ex){
      ex.printStackTrace();
    }
  }

  /**
   * Processes an instruction by sending it to all clients.
   *
   * @param handler IOHandler
   * @param clientNumber int
   * @param input DataInputStream
   * @param output DataOutputStream
   * @throws IOException
   */
  protected synchronized void processInstruction(IOHandler handler, int clientNumber,
      DataInputStream input, DataOutputStream output) throws IOException{
    splitOutput.writeByte(handler.uniqueId);
    splitOutput.writeInt(getTimeStamp());
    handler.echo(clientNumber, input, splitOutput);
  }

  /**
   * An object of this class is used to accept connecting clients by means
   * of a ServerSocket.
   */
  private class ListeningThread
      extends Thread{

    public boolean isListening = true;

    public ListeningThread(){
      super("Server Listening Thread");
    }

    /**
     * Accepts new clients until the stopListening method is invoked.
     */
    public void run(){
      System.out.println("Server: Now listening for clients, using port " +
                         servSock.getLocalPort());
      while(isListening){
        try{
          Socket newSocket = servSock.accept();
          System.out.println("Server: Client connection accepted to " +
                             newSocket);
          synchronized(this){
            addClient(new ClientThread(newSocket));
          }
        } catch(IOException ex){
          if(!servSock.isClosed()){
            System.out.println(
                "Server: Unexpected Exception in Listening Thread!");
            ex.printStackTrace();
          }
        }
      }
      System.out.println("Server: Listening Thread terminated.");
    }

  }

  /**
   * This class represents clients connected to this Server.
   */
  class ClientThread
      extends Thread{

    private final DataInputStream input;
    private final DataOutputStream output;
    private final int clientNumber;
    private boolean isListening = true;

    public ClientThread(Socket socket) throws IOException{
      super("Server Link Thread");
      this.clientNumber = nextClientNumber++;
      this.input = new DataInputStream(new BufferedInputStream(socket.
          getInputStream()));
      this.output = new DataOutputStream(new BufferedOutputStream(socket.
          getOutputStream()));
      output.writeInt(clientNumber);
      output.flush();
    }

    /**
     * Listens for instructions continually until an exception is thrown,
     * then terminates.
     */
    public void run(){
      try{
        while(isListening){
          byte type = input.readByte();
          IOHandler handler = protocol.getType(type);
          synchronized(Server.this){
            processInstruction(handler, clientNumber, input, splitOutput);
            splitOutput.flush();
          }
        }
      } catch(IOException ex){
        reportClientException(this, ex);
      } finally{
        System.out.println("Server: Client listening thread terminated");
      }
    }

    public String toString(){
      return super.toString() + " " + clientNumber;
    }

  }

  /**
   * This OutputStream will write data to all client OutputStreams,
   * thus distributing the data to all clients.
   */
  private class SplitStream
      extends OutputStream{

    public void write(int b){
      for(ClientThread c : clients){
        try{
          c.output.writeByte(b);
        } catch(IOException ex){
          reportClientException(c, ex);
        }
      }
    }

    public void flush(){
      for(ClientThread c : clients){
        try{
          c.output.flush();
        } catch(IOException ex){
          ex.printStackTrace();
          reportClientException(c, ex);
        }
      }
    }

  }

}
