package jwars.test.explosions;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import java.awt.AlphaComposite;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.Image;
import java.awt.Transparency;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.Timer;

import jwars.Util;

public class ExplosionGenerator {

    private final GraphicsConfiguration gc =
            java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment().
            getDefaultScreenDevice().getDefaultConfiguration();

    private final Image basic;
    private final int
            minSparkRadius, maxSparkRadius,
            minSparkDuration, maxSparkDuration;

    public ExplosionGenerator(int minSparkRadius, int maxSparkRadius, int minSparkDuration, int maxSparkDuration){
        this.minSparkRadius = minSparkRadius;
        this.maxSparkRadius = maxSparkRadius;
        this.minSparkDuration = minSparkDuration;
        this.maxSparkDuration = maxSparkDuration;
        Image im = null;
        try {
          im = Util.readImage("basicexplosion.png", Transparency.TRANSLUCENT);
        } catch (Exception ex) {
            ex.printStackTrace();
            System.exit(0);
        }
        basic = im;
    }

    public List<Image> createRoundExplosion(int spread, int sparksPerFrame, int duration, long seed){
        ArrayList<Image> frames = new ArrayList<Image>();
        ArrayList<Spark> sparks = new ArrayList<Spark>();
        Random rand = new Random(seed);
        //Continue until all sparks are gone
        for(int time = 0; sparks.size() > 0 || time == 0; time++){
            //Create some sparks unless we have passed the duration
            if(time < duration){
                for (int i = 0; i < sparksPerFrame; i++) {
                    int sparkDuration = nextInt(rand, minSparkDuration,
                                                maxSparkDuration);
                    int size = nextInt(rand, minSparkRadius, maxSparkRadius);
                    double angle = rand.nextFloat() * 2 * Math.PI;
                    int dist = rand.nextInt(spread);
                    int x = (int) (dist * Math.cos(angle));
                    int y = (int) (dist * Math.sin(angle));
                    sparks.add(new Spark(sparkDuration, size, x, y));
                }
            }
            int size = getMaxSize(sparks);
            Image frame = gc.createCompatibleImage(size, size, Transparency.TRANSLUCENT);
            Graphics g = frame.getGraphics();
            g.translate(size/2, size/2);
            render(sparks, g);
            g.dispose();
            frames.add(frame);
            //displayImage(frame);
        }
        return frames;
    }

    private int getMaxSize(List<Spark> sparks){
        int maxRadius = 0;
        int currentRadius = 0;
        for(Spark s : sparks){
            currentRadius = Math.max(s.getMaxHorizontalDistFromZero(),
                                     s.getMaxVerticalDistFromZero());
            maxRadius = Math.max(maxRadius, currentRadius);
        }
        return 2*maxRadius;
    }

    private int nextInt(Random rand, int min, int max){
        return min + rand.nextInt(max-min);
    }

    private void render(ArrayList<Spark> sparks, Graphics g){
        for(int i = 0; i < sparks.size(); i++){
            boolean hasMoreElements = sparks.get(i).drawNextFrame(g);
            if(!hasMoreElements){
                sparks.remove(i);
                i--;
            }
        }
    }

    public static void main(String[] args){
        JFrame frame = new JFrame();

        if(args.length == 1 && args[0].equals("-h")){
          System.out.println("Parameters: ");
          System.out.println("maxSize");
          System.out.println("minSize");
          System.out.println("minTime");
          System.out.println("maxTime");
          System.out.println("spread");
          System.out.println("sparksPerFrame");
          System.out.println("duration");
          System.out.println("seed");
          System.exit(0);
        }

        int minSize = 8;
        int maxSize = 20;
        int minTime = 5;
        int maxTime = 10;

        int spread = 8;
        int sparksPerFrame = 16;
        int duration = 5;
        long seed = 0;

        if(args.length == 8){
          minSize = Integer.parseInt(args[0]);
          maxSize = Integer.parseInt(args[1]);
          minTime = Integer.parseInt(args[2]);
          maxTime = Integer.parseInt(args[3]);
          spread  = Integer.parseInt(args[4]);
          sparksPerFrame = Integer.parseInt(args[5]);
          duration = Integer.parseInt(args[6]);
          seed = Integer.parseInt(args[7]);
        }

        ExplosionGenerator expl = new ExplosionGenerator(minSize, maxSize, minTime, maxTime);
        List<Image> images = expl.createRoundExplosion(spread, sparksPerFrame, duration, seed);
        RenderingPanel panel = new RenderingPanel(images);
        frame.add(panel);
        Timer timer = new Timer(100, panel);
        panel.setPreferredSize(new Dimension(800,600));
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.pack();
        frame.setVisible(true);
        timer.start();
    }

    private static void displayImage(Image im){
        JFrame frame = new JFrame();
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        frame.add(new JLabel(new ImageIcon(im)));
        frame.pack();
        frame.setVisible(true);
    }

    private static class RenderingPanel extends JPanel implements ActionListener {

        int time;
        List<Image> frames;

        public RenderingPanel(List<Image> frames){
            this.frames = frames;
        }

        protected void paintComponent(Graphics g){
            super.paintComponent(g);
            Image im = frames.get(time % frames.size());
            int cx = this.getWidth()/2;
            int cy = this.getHeight()/2;
            int x = cx - im.getWidth(null)/2;
            int y = cy - im.getHeight(null)/2;
            g.drawImage(im, x, y, null);
        }

        public void actionPerformed(ActionEvent actionEvent) {
            time++;
            repaint();
        }

    }

    //private final int maxSize = 24, minSize = 6;

    private static float getSize(float T, float M, float t){
        return 4/T/T*M*(T-t)*t;
    }

    private class Spark {

        Image[] frames;
        //int creationTime;
        int maxTime;
        int x, y;
        int radius;
        int time;

        //drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, Color bgcolor, ImageObserver observer)
        public Spark(int frameCount, int size, int x, int y){
            frames = new Image[frameCount];
            //this.creationTime = ExplosionGenerator.this.time;
            this.x = x;
            this.y = y;
            maxTime = frameCount;
            for(int i = 0; i < frameCount; i++){
                int imageSize = (int)getSize(frameCount, size, i);
                if(imageSize < 1)
                    imageSize = 1;
                frames[i] = gc.createCompatibleImage(imageSize, imageSize, Transparency.TRANSLUCENT);//new BufferedImage(imageSize, imageSize, Transparency.TRANSLUCENT);
                Graphics2D g = (Graphics2D)frames[i].getGraphics();
                int dx1 = 0, dx2 = frames[i].getWidth(null), dy1 = 0, dy2 = frames[i].getHeight(null),
                        sx1 = 0, sx2 = basic.getWidth(null), sy1 = 0, sy2 = basic.getHeight(null);
                //if(i > maxTime/2){
                g.setComposite(AlphaComposite.getInstance(AlphaComposite.
                        SRC_OVER, getAlpha(i)));
                //}
                g.drawImage(basic, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, null, null);
                g.dispose();
            }
        }

        public int getMaxHorizontalDistFromZero(){
            if(time >= frames.length){
                throw new RuntimeException("Explosion already over!");
            }
            return frames[time].getWidth(null) + Math.abs(x);
        }

        public int getMaxVerticalDistFromZero(){
            if(time >= frames.length){
                throw new RuntimeException("Explosion already over!");
            }
            return frames[time].getHeight(null) + Math.abs(y);
        }

        private float getAlpha(int time){
            float baseTime = 0;
            float slope = 1f/(baseTime - maxTime);
            float intersect = - slope * maxTime;
            return (float)Math.pow(slope * time + intersect,2);
        }

        public boolean drawNextFrame(Graphics g){
            time++;
            if(time >= frames.length){
                return false;
            } else {
                Image i = frames[time];
                int w = i.getWidth(null);
                int h = i.getHeight(null);
                /*if(time > maxTime/2){
                    Graphics2D g2d = (Graphics2D)g;
                    g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getAlpha()));
                }*/
                g.drawImage(i, x-w/2, y-h/2, null);
                return true;
            }
        }

    }

}
